#!/usr/bin/python
#    "$Name:  $";
#    "$Header:  $";
#=============================================================================
#
# file :        GroupedAttributesManager.py
#
# description : Python source for the GroupedAttributesManager and its commands. 
#                The class is derived from Device. It represents the
#                CORBA servant object which will be accessed from the
#                network. All commands which can be executed on the
#                GroupedAttributesManager are implemented in this file.
#
# project :     TANGO Device Server
#
# $Author:  $
#
# $Revision:  $
#
# $Log:  $
#
# copyleft :    European Synchrotron Radiation Facility
#               BP 220, Grenoble 38043
#               FRANCE
#
#=============================================================================
#          This file is generated by POGO
#    (Program Obviously used to Generate tango Object)
#
#         (c) - Software Engineering Group - ESRF
#=============================================================================
#

import PyTango
import sys
import numpy as np
import time
from PyTango._PyTango import CmdArgType

MAX_POINTS = 500
NUM_ATTRIBUTES = 30
#==================================================================
#   GroupedAttributesManager Class Description:
#
#         High level management of the LDM power supplies
#
#==================================================================
#     Device States Description:
#
#   DevState.ON :
#   DevState.FAULT :
#==================================================================


class GroupedAttributesManager(PyTango.Device_4Impl):

#--------- Add you global variables here --------------------------

#------------------------------------------------------------------
#    Device constructor
#------------------------------------------------------------------
    def __init__(self,cl, name):
        PyTango.Device_4Impl.__init__(self,cl,name)

        self.attr_proxies = []
        self.attr_StoredGroupNames = []
        self.attr_StoredGroups =  []
        self.attr_StoredGroupsAsString = ""
        self.MYSELF_Proxy = None
	
        GroupedAttributesManager.init_device(self)

#------------------------------------------------------------------
#    Device destructor
#------------------------------------------------------------------
    def delete_device(self):
        #print "[Device delete_device method] for device",self.get_name()
	return


#------------------------------------------------------------------
#    Device initialization
#------------------------------------------------------------------
    def init_device(self):
        #print "In ", self.get_name(), "::init_device()"
        self.set_state(PyTango.DevState.ON)
        self.get_device_properties(self.get_device_class())
        del self.attr_proxies[:]
        if self.MYSELF_Proxy == None:
            self.MYSELF_Proxy= PyTango.DeviceProxy(self.get_name())

        for attrname in self.attr_names:
            try:
                self.attr_proxies.append( PyTango.AttributeProxy(attrname) )
            except PyTango.DevFailed, ex:
                self.status = "attribute " + attrname + " Proxy error"
                #self.logger.error(self.status)
                self.set_state(PyTango.DevState.FAULT)
                PyTango.Except.print_exception(ex)
                return

#------------------------------------------------------------------
#    Always excuted hook method
#------------------------------------------------------------------
    def always_executed_hook(self):
        #print "In ", self.get_name(), "::always_excuted_hook()"
        if self.get_state() == PyTango.DevState.FAULT:
            self.init_device()
        if self.get_state() == PyTango.DevState.FAULT:
            PyTango.Except.throw_exception("Device in Fault state","Check fault proxy!","always_executed_hook")


#==================================================================
#
#    GroupedAttributesManager read/write attribute methods
#
#==================================================================
#------------------------------------------------------------------
#    Read Attribute Hardware
#------------------------------------------------------------------
    def read_attr_hardware(self,data):
        #print "In ", self.get_name(), "::read_attr_hardware()"
	return



#------------------------------------------------------------------
#    Read StoredGroupNames attribute
#------------------------------------------------------------------
    def read_StoredGroupNames(self, attr):
        #print "In ", self.get_name(), "::read_StoredGroupNames()"
        
        #    Add your own code here
        attr.set_value(self.attr_StoredGroupNames)



#------------------------------------------------------------------
#    Read StoredGroups attribute
#------------------------------------------------------------------
    def read_StoredGroups(self, attr):
        #print "In ", self.get_name(), "::read_StoredGroups()"
        
        #    Add your own code here
        attr.set_value(self.attr_StoredGroups)


#------------------------------------------------------------------
#    Read StoredGroupsAsString attribute
#------------------------------------------------------------------
    def read_StoredGroupsAsString(self, attr):
        #print "In ", self.get_name(), "::read_StoredGroupsAsString()"
        
        #    Add your own code here
        attr.set_value(self.attr_StoredGroupsAsString)


#------------------------------------------------------------------
#    Write StoredGroupsAsString attribute
#------------------------------------------------------------------
    def write_StoredGroupsAsString(self, attr):
        #print "In ", self.get_name(), "::write_StoredGroupsAsString()"

        #    Add your own code here
        self.attr_StoredGroupsAsString  = attr.get_write_value()
        del self.attr_StoredGroupNames[:]
        del self.attr_StoredGroups[:]
        for line in self.attr_StoredGroupsAsString.split(";"):
            if line == '':
                continue
            pos_data = line.split(",")
            self.attr_StoredGroupNames.append(pos_data[0])
            pos_data.pop(0)
            tmp_str = ""
            self.attr_StoredGroups.append(pos_data)



#==================================================================
#
#    GroupedAttributesManager command methods
#
#==================================================================

#------------------------------------------------------------------
#    StoreActualGroup command:
#
#    Description: 
#    argin:  DevString    Name that will be assigned to this stored attribute
#------------------------------------------------------------------
    def StoreActualGroupValues(self, argin):
        #print "In ", self.get_name(), "::StoreActualGroup()"
        #    Add your own code here
        tmp_lst = []
	try:
            for proxy in self.attr_proxies:
	    	if self.use_write_values:
			value=proxy.read().w_value
		else:
			value=proxy.read().value
		tmp_lst.append(str(value))
        except PyTango.DevFailed, ex:
            self.status = "Proxy error"
            self.logger.error(self.status)
            self.set_state(PyTango.DevState.FAULT)
            PyTango.Except.throw_exception("Device in Fault state","Check fault proxy!","StoreActualGroup()")
	found = False
        for idx in range(len(self.attr_StoredGroupNames)):
            if (self.attr_StoredGroupNames[idx] == argin):
                found = True
                self.attr_StoredGroups[idx] = tmp_lst
                break
        if (not found):
            self.attr_StoredGroups.append(tmp_lst)
            self.attr_StoredGroupNames.append(argin)
	    
        tmsString = ""
        for idx in range(len(self.attr_StoredGroupNames)):
            tmsString += self.attr_StoredGroupNames[idx]
	    for item in self.attr_StoredGroups[idx]:
	            tmsString += ","
        	    tmsString += item
            tmsString += ";"
        self.MYSELF_Proxy.write_attribute("StoredGroupsAsString",tmsString)


#------------------------------------------------------------------
#    MoveToGroupByName command:
#
#    Description: 
#    argin:  DevString    Name of the attribute where to move
#------------------------------------------------------------------
    def ApplyGroupValuesByName(self, argin):
        #print "In ", self.get_name(), "::MoveToGroupByName()"
        #    Add your own code here
        try:
            for idx in range(len(self.attr_StoredGroupNames)):
                if (self.attr_StoredGroupNames[idx] == argin):
                    ii=0
                    for value in self.attr_StoredGroups[idx]:
                        dtype = self.attr_proxies[ii].get_config().data_type
                        if dtype in [CmdArgType.DevFloat, CmdArgType.DevDouble]:
                            self.attr_proxies[ii].write(float(value))
                        elif dtype in [CmdArgType.DevBoolean]:
                            self.attr_proxies[ii].write(bool(value))
                        elif dtype in [CmdArgType.DevString]:
                            self.attr_proxies[ii].write(str(value))
                        else:
                            self.attr_proxies[ii].write(int(value))
                        time.sleep(self.apply_delay_msec / 1000.)
                        ii = ii + 1
                    return
        except PyTango.DevFailed, ex:
            self.status = "Proxy error"
            self.set_state(PyTango.DevState.FAULT)
            self.set_status(str(ex))
            PyTango.Except.throw_exception("Device in Fault state","Check fault proxy!","ApplyGroupValuesByName()")
        # Group name not found
        PyTango.Except.throw_exception("Group name not found","Check input group name!","ApplyGroupValuesByName()")
        
       
#------------------------------------------------------------------
#    MoveToGroupByIndex command:
#
#    Description: 
#    argin:  DevShort    Index of the attribute where to move
#------------------------------------------------------------------
    def ApplyGroupValuesByIndex(self, argin):
        #print "In ", self.get_name(), "::MoveToGroupByName()"
        #    Add your own code here
        if (argin >= len(self.attr_StoredGroups) or argin < 0):
            PyTango.Except.throw_exception("Input index out of range","Input index out of range!","ApplyGroupValuesByIndex()")
        try:
            if (argin < len(self.attr_StoredGroups)):
                idx = argin
                ii=0
                for value in self.attr_StoredGroups[idx]:
                    dtype = self.attr_proxies[ii].get_config().data_type
                    if dtype in [CmdArgType.DevFloat, CmdArgType.DevDouble]:
                        self.attr_proxies[ii].write(float(value))
                    elif dtype in [CmdArgType.DevBoolean]:
                        self.attr_proxies[ii].write(bool(value))
                    elif dtype in [CmdArgType.DevString]:
                        self.attr_proxies[ii].write(str(value))
                    else:
                        self.attr_proxies[ii].write(int(value))
                    time.sleep(self.apply_delay_msec / 1000.)
                    ii = ii + 1
        except PyTango.DevFailed, ex:
            self.status = "Proxy error"
            self.logger.error(self.status)
            self.set_state(PyTango.DevState.FAULT)
            PyTango.Except.throw_exception("Device in Fault state","Check fault proxy!","ApplyGroupValuesByIndex()")

#------------------------------------------------------------------
#    RemoveGroupByName command:
#
#    Description: 
#    argin:  DevString    Name of the attribute to remove
#------------------------------------------------------------------
    def RemoveGroupByName(self, argin):
        #print "In ", self.get_name(), "::RemoveGroupByName()"
        #    Add your own code here
        if argin in self.protected_group_names:
            PyTango.Except.throw_exception("Mandatory Group name","Group cannot be removed!","RemoveGroupByName()")
        found = False
        for idx in range(len(self.attr_StoredGroupNames)):
            if (self.attr_StoredGroupNames[idx] == argin):
                found = True
                self.attr_StoredGroups.pop(idx)
                self.attr_StoredGroupNames.pop(idx)
                break
        if (not found):
            PyTango.Except.throw_exception("Group name not found","Check the input name!","RemoveGroupByName()")

        tmsString = ""
        for idx in range(len(self.attr_StoredGroupNames)):
            tmsString += self.attr_StoredGroupNames[idx]
	    for value in self.attr_StoredGroups[idx]:
	    	tmsString += ","
            	tmsString += value
            tmsString += ";"
        self.MYSELF_Proxy.write_attribute("StoredGroupsAsString",tmsString)
       
#------------------------------------------------------------------
#    RemoveGroupByIndex command:
#
#    Description: 
#    argin:  DevShort    Index of the attribute to remove
#------------------------------------------------------------------
    def RemoveGroupByIndex(self, argin):
        #print "In ", self.get_name(), "::RemoveGroupByIndex()"
        #    Add your own code here
        if (argin >= len(self.attr_StoredGroups) or argin < 0):
            PyTango.Except.throw_exception("Index out of range","Check the input index!","RemoveGroupByIndex()")

        if self.attr_StoredGroupNames[argin] in self.protected_group_names:
            PyTango.Except.throw_exception("Mandatory Group name","Group cannot be removed!","RemoveGroupByIndex()")
        self.attr_StoredGroups.pop(argin)
        self.attr_StoredGroupNames.pop(argin)

        tmsString = ""
        for idx in range(len(self.attr_StoredGroupNames)):
            tmsString += self.attr_StoredGroupNames[idx]
	    for value in self.attr_StoredGroups[idx]:
	    	tmsString += ","
            	tmsString += value
            tmsString += ";"
        self.MYSELF_Proxy.write_attribute("StoredGroupsAsString",tmsString)


#==================================================================
#
#    GroupedAttributesManagerClass class definition
#
#==================================================================
class GroupedAttributesManagerClass(PyTango.DeviceClass):

    #    Class Properties
    class_property_list = {
        }


    #    Device Properties
    device_property_list = {
        'attr_names':
            [PyTango.DevVarStringArray,
            "Attribute names belonging to the group",
            [ "ldm/hv_supply/attributeA" ] ],
        'protected_group_names':
            [PyTango.DevVarStringArray,
            "List of group names that should exist",
            [ ] ],
        'apply_delay_msec':
            [PyTango.DevShort,
            "Delay between apply actions",
            [ "100" ] ],
        'use_write_values':
            [PyTango.DevBoolean,
            "Delay between apply actions",
            [ "0" ] ],
        }


    #    Command definitions
    cmd_list = {
        'StoreActualGroupValues':
            [[PyTango.DevString, "Name that will be assigned to this stored group of values"],
            [PyTango.DevVoid, ""]],
        'ApplyGroupValuesByName':
            [[PyTango.DevString, "Name of the group where to apply"],
            [PyTango.DevVoid, ""]],
        'ApplyGroupValuesByIndex':
            [[PyTango.DevShort, "Index of the group where to apply"],
            [PyTango.DevVoid, ""]],
        'RemoveGroupByName':
            [[PyTango.DevString, "Name of the group to remove"],
            [PyTango.DevVoid, ""]],
        'RemoveGroupByIndex':
            [[PyTango.DevShort, "Index of the group to remove"],
            [PyTango.DevVoid, ""]],
        }


    #    Attribute definitions
    attr_list = {
        'StoredGroupNames':
            [[PyTango.DevString,
            PyTango.SPECTRUM,
            PyTango.READ, MAX_POINTS],
            {
                'description':"Array of stored group names",
            } ],
        'StoredGroups':
            [[PyTango.DevString,
            PyTango.IMAGE,
            PyTango.READ, NUM_ATTRIBUTES, MAX_POINTS],
            {
                'description':"2D Array of stored values",
            } ],
        'StoredGroupsAsString':
            [[PyTango.DevString,
            PyTango.SCALAR,
            PyTango.READ_WRITE],
            {
                'description':"Stored values memorized as a string\nFormat of the string: 'NAME1,V1,V2,V3,V4,V5,..;NAME2,V1,V2,V3,V4,V5,..",
                'Memorized':"true",
            } ],
        }


#------------------------------------------------------------------
#    GroupedAttributesManagerClass Constructor
#------------------------------------------------------------------
    def __init__(self, name):
        PyTango.DeviceClass.__init__(self, name)
        self.set_type(name);
        #print "In GroupedAttributesManagerClass  constructor"

#==================================================================
#
#    GroupedAttributesManager class main method
#
#==================================================================
def main():
    try:
        py = PyTango.Util(sys.argv)
        py.add_TgClass(GroupedAttributesManagerClass,GroupedAttributesManager,'GroupedAttributesManager')

        U = PyTango.Util.instance()
        U.server_init()
        U.server_run()

    except PyTango.DevFailed,e:
        print '-------> Received a DevFailed exception:',e
    except Exception,e:
        print '-------> An unforeseen exception occured....',e

if __name__ == '__main__':
    main()
